/*
 *  Simple test program to demonstrate strict order mutex
 *
 */

#define CONFIGURE_INIT

#include <bsp.h>
#include <stdio.h>
#include "tmacros.h"


#define BACK_TYPE(_type_in_ptr,_type_out,_type_in_name)		\
  ((_type_out *)((unsigned int)_type_in_ptr - (unsigned int)(&((_type_out *)0)->_type_in_name)))


/* configuration information */

#define CONFIGURE_APPLICATION_NEEDS_CONSOLE_DRIVER
#define CONFIGURE_APPLICATION_NEEDS_CLOCK_DRIVER

#define CONFIGURE_RTEMS_INIT_TASKS_TABLE

#define CONFIGURE_EXTRA_TASK_STACKS (RTEMS_MINIMUM_STACK_SIZE * 3)

#define CONFIGURE_MAXIMUM_TASKS 10
#define CONFIGURE_MAXIMUM_SEMAPHORES 10


void p_mutex_chain(Thread_Control *the_thread);
rtems_task Task0(rtems_task_argument ignored);
rtems_task Task1(rtems_task_argument ignored);
rtems_task Init(rtems_task_argument ignored);


#include <rtems/confdefs.h>



rtems_id Task_id[4];
rtems_name Task_name[4];

rtems_id Mutex_id[4];
rtems_name Mutex_name[4];

rtems_task Init(rtems_task_argument ignored)
{
  rtems_status_code status;
  printf("\n----------------TEST 36 ------------\n");

  Mutex_name[0] = rtems_build_name( 'S','0',' ',' ');
  status = rtems_semaphore_create(
				  Mutex_name[0],
				  1,
				  RTEMS_LOCAL|
				  RTEMS_SIMPLE_BINARY_SEMAPHORE|
				  RTEMS_INHERIT_PRIORITY|
				  RTEMS_PRIORITY,
				  0,
				  &Mutex_id[0]);

  Mutex_name[1] = rtems_build_name( 'S','1',' ',' ');
  status = rtems_semaphore_create(
				  Mutex_name[1],
				  1,
				  RTEMS_LOCAL|
				  RTEMS_SIMPLE_BINARY_SEMAPHORE|
				  RTEMS_PRIORITY_CEILING|
				  RTEMS_PRIORITY,
				  2,
				  &Mutex_id[1]);
  Mutex_name[2] = rtems_build_name( 'S','Y','N','C');

				 
  Task_name[0] = rtems_build_name( 'T','0',' ',' ');
  status = rtems_task_create(
			     Task_name[0],
			     4,
			     RTEMS_MINIMUM_STACK_SIZE *2,
			     RTEMS_DEFAULT_MODES,
			     RTEMS_DEFAULT_ATTRIBUTES,
			     &Task_id[0]
			     );

  printf("Id of task0 is %d\n",Task_id[0]);

  status = rtems_task_start( Task_id[0],Task0, 0);

  status = rtems_task_delete( RTEMS_SELF);
}


rtems_task Task0(rtems_task_argument ignored)
{
  rtems_status_code status;
  
  status = rtems_semaphore_obtain(
				  Mutex_id[0],
				  RTEMS_WAIT,
				  0);
  printf("I have got Mutex0\n");
 
  status = rtems_semaphore_obtain(
				  Mutex_id[1],
				  RTEMS_WAIT,
				  0);

  printf("I have got Mutex1,and my priority is set to the ceiling of S1\n");
  p_mutex_chain(_Thread_Executing);

  status = rtems_semaphore_release( Mutex_id[0] );
  if(status == CORE_MUTEX_RELEASE_NOT_ORDER)
    printf("\n**Error,when task0 releasing S0 not in order\n");

  rtems_semaphore_release(Mutex_id[1]);
  printf("after releasing mutex 1\n");
  p_mutex_chain(_Thread_Executing);

  Task_name[1] = rtems_build_name( 'T','1',' ',' ');
  status = rtems_task_create(
			     Task_name[1],
			     1,
			     RTEMS_MINIMUM_STACK_SIZE *2,
			     RTEMS_DEFAULT_MODES,
			     RTEMS_DEFAULT_ATTRIBUTES,
			     &Task_id[1]
			     );
  printf("Id of task 1 is %d\n",Task_id[1]);

  status = rtems_task_start( Task_id[1],Task1, 0);

  printf("I am task2,my priority is improved\n");
  p_mutex_chain(_Thread_Executing);

  rtems_semaphore_release(Mutex_id[0]);
  
  while(1);
}
  

rtems_task Task1(rtems_task_argument ignored)
{
  rtems_semaphore_obtain(
			 Mutex_id[0],
			 RTEMS_WAIT,
			 0);

  printf("I am task1, I have got S0\n");
  p_mutex_chain(_Thread_Executing);
  while(1);
}
	 

void p_mutex_chain(Thread_Control *the_thread)
{
  Chain_Control *con = &the_thread->lock_mutex;
  Chain_Node  * node = con->first;
  CORE_mutex_Control * p_mutex ;

  if(!_Chain_Is_empty(&the_thread->lock_mutex)){
    while(node != _Chain_Tail(con)){
      p_mutex = BACK_TYPE(node,CORE_mutex_Control,queue);
      printf("node:Id=%p,priority_before=%d,"
	     "holder_id=%d,holder's current priority=%d\n",node,
	     ((CORE_mutex_order_list *)node)->priority_before,p_mutex->holder_id,p_mutex->holder->current_priority);
      node = node->next;    
    }
  }
  else
    printf("the Chain is empty\n");
}
    
